import os
from datetime import date

import factory
import pytest
from factory.alchemy import SQLAlchemyModelFactory
from flask_bcrypt import generate_password_hash
from pytest_factoryboy import register
from sqlalchemy import create_engine, event

from alembic import command
from alembic.config import Config as AlembicConfig
from pychores import create_app
from pychores.configmodule import Config
from pychores.model import Chore, DeferedSession, Task, User


def pytest_addoption(parser):
    parser.addoption("--build", action="store_true", help="test app contains files")


def pytest_runtest_setup(item):
    if "build" in item.keywords and not item.config.getvalue("build"):
        pytest.skip("need --build option to run")


SessionLocal = DeferedSession.get_session_local(
    Config.get_config("test").SQLALCHEMY_DATABASE_URI
)


class BaseFactory(SQLAlchemyModelFactory):
    class Meta:
        abstract = True
        sqlalchemy_session = SessionLocal
        sqlalchemy_session_persistence = "commit"


@pytest.fixture(scope="session")
def connection():
    """
    recreate the database for each test :(
    i'm not smart enough to make the rollback thing works properly
    """
    ini_location = os.path.join(os.path.dirname(__file__), "..", "alembic.ini")
    sqlalchemy_url = Config.get_config("test").SQLALCHEMY_DATABASE_URI
    alembic_config = AlembicConfig(ini_location)
    alembic_config.set_main_option("sqlalchemy.url", sqlalchemy_url)
    command.upgrade(alembic_config, "head")
    engine = create_engine(Config.get_config("test").SQLALCHEMY_DATABASE_URI)
    connection = engine.connect()
    yield connection
    connection.close()
    command.downgrade(alembic_config, "base")


@pytest.fixture
def app(connection):
    app = create_app("test")
    yield app
    app.db_session.remove()


@pytest.fixture()
def db_session(connection):
    transaction = connection.begin()
    session = SessionLocal(bind=connection)
    session.begin_nested()

    @event.listens_for(session, "after_transaction_end")
    def restart_savepoint(db_session, transaction):
        if transaction.nested and not transaction._parent.nested:
            session.expire_all()
            session.begin_nested()

    yield session

    SessionLocal.remove()
    transaction.rollback()


@register
class UserFactory(BaseFactory):
    class Meta:
        model = User

    username = factory.Sequence(lambda n: f"user_{n}")
    password = factory.Sequence(
        lambda n: generate_password_hash(f"password_{n}").decode("utf-8")
    )
    email = factory.Sequence(lambda n: f"user_{n}@example.com")


@register
class ChoreFactory(BaseFactory):
    class Meta:
        model = Chore

    name = factory.Sequence(lambda n: f"chore_{n}")
    description = factory.Sequence(lambda n: f"description of chore_{n}")
    period = 5
    user = factory.SubFactory(UserFactory)


@register
class TaskFactory(BaseFactory):
    class Meta:
        model = Task

    execution_date = date(year=2020, month=5, day=18)
    chore = factory.SubFactory(ChoreFactory)


@pytest.fixture
def headers():
    mimetype = "application/json"
    headers = {
        "Content-Type": mimetype,
        "Accept": mimetype,
    }
    return headers



"""Add email in user table

Revision ID: 944acee085dd
Revises: 4f016505c7c3
Create Date: 2020-07-28 18:15:44.522287

"""
import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision = "944acee085dd"
down_revision = "4f016505c7c3"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "user",
        sa.Column(
            "email",
            sa.String(length=50),
            nullable=False,
            server_default="test@example.com",
        ),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("user", "email")
    # ### end Alembic commands ###



from abc import ABC, abstractmethod

from pychores.domain.entity.chore import Chore


class IGetChore(ABC):
    @abstractmethod
    def get_chore(self, username: str, chore_id: int) -> Chore:
        """Return chore owned by username"""


class GetChore:
    def __init__(self, repo: IGetChore):
        self.repo = repo

    def execute(self, username: str, chore_id: int) -> Chore:
        return self.repo.get_chore(username, chore_id)



from abc import ABC, abstractmethod

from pychores.domain.entity.task import Task


class IGetTask(ABC):
    @abstractmethod
    def get_task(self, username: str, task_id: int) -> Task:
        """Get task owned by username"""


class GetTask:
    def __init__(self, repo: IGetTask):
        self.repo = repo

    def execute(self, username: str, task_id: int):
        task = self.repo.get_task(username=username, task_id=task_id)
        return task



from flask import Blueprint, current_app, jsonify, request, url_for
from flask_cors import CORS
from flask_jwt_extended import get_jwt_identity, jwt_required

from pychores.adapter.repository.sqla.task import TaskRepo
from pychores.domain.use_cases.delete_task import DeleteTask
from pychores.domain.use_cases.get_task import GetTask

from .. import respond_success

bp = Blueprint("tasks", __name__, url_prefix="/api")

CORS(bp, resources={r"/*": {"origins": "*"}})


def serialize_task(task):
    return {
        "id": task.id,
        "chore_id": task.chore.id,
        "chore_name": task.chore.name,
        "execution_date": task.execution_date,
    }


def respond_task(task):
    response = jsonify(serialize_task(task))
    response.headers["Location"] = url_for("tasks.task", task_id=task.id)
    return response


@bp.route("/task/<int:task_id>", methods=["GET", "DELETE"])
@jwt_required()
def task(task_id):
    username = get_jwt_identity()
    if request.method == "DELETE":
        uc = DeleteTask(TaskRepo(current_app.db_session))
        uc.execute(username, task_id)
        return "", 204
    # get
    uc = GetTask(TaskRepo(current_app.db_session))
    task = uc.execute(username=username, task_id=task_id)
    return respond_success("task", serialize_task(task))



import os

DEFAULT_SECRET = "do not tell anyone this is a secret"


class Config:
    DEBUG = False
    TESTING = False
    SQLALCHEMY_DATABASE_URI = os.getenv("DATABASE_URL")
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    SECRET_KEY = os.getenv("FLASK_SECRET") or DEFAULT_SECRET
    PG_HOST = os.getenv("POSTGRES_HOST") or "localhost"
    JWT_ACCESS_TOKEN_EXPIRES = False

    @staticmethod
    def get_config(env):
        if env == "development":
            return Development()
        if env == "production":
            return Production()
        if env == "test":
            return Test()
        return Production()


class Production(Config):
    def __init__(self):
        if self.SQLALCHEMY_DATABASE_URI is None:
            raise ValueError(
                "Database must be set in production mode, set DATABASE_URL env var"
            )
        if self.SECRET_KEY == DEFAULT_SECRET:
            raise ValueError(
                "Secret key can't be default in production, set FLASK_SECRET env var"
            )


class Development(Config):
    DEBUG = True
    ENV = True

    def __init__(self):
        if self.SQLALCHEMY_DATABASE_URI is None:
            self.SQLALCHEMY_DATABASE_URI = "sqlite:///chores.db"


class Test(Config):
    DEBUG = True
    ENV = True

    def __init__(self):
        self.SQLALCHEMY_DATABASE_URI = (
            f"postgresql://chores@{self.PG_HOST}/kinky_chores_tests"
        )



"""Init database

Revision ID: 615f2c3df9bd
Revises:
Create Date: 2020-05-02 13:59:35.346971

"""
import sqlalchemy as sa
from sqlalchemy.sql import column, table

from alembic import context, op

# revision identifiers, used by Alembic.
revision = "615f2c3df9bd"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    schema_upgrade()
    if context.get_x_argument(as_dictionary=True).get("data"):
        data_upgrade()


def downgrade():
    if context.get_x_argument(as_dictionary=True).get("data"):
        data_downgrade()
    schema_downgrade()


def schema_upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "chore",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("description", sa.String(), nullable=False),
        sa.Column("period", sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("name"),
    )
    op.create_table(
        "task",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("chore_id", sa.Integer(), nullable=False),
        sa.Column("execution_date", sa.Date(), nullable=False),
        sa.ForeignKeyConstraint(
            ["chore_id"],
            ["chore.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def schema_downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("task")
    op.drop_table("chore")
    # ### end Alembic commands ###


def data_upgrade():
    chore_table = table(
        "chore",
        column("name", sa.String()),
        column("description", sa.String()),
        column("period", sa.Integer()),
    )
    op.bulk_insert(
        chore_table,
        [
            {
                "name": "Sanitaire",
                "description": "Nettoyer les vasques de la salle de bain et des toilettes",
                "period": 7,
            },
            {
                "name": "Sols",
                "description": "Nettoyer les sols de la cuisine et de la salle à manger",
                "period": 14,
            },
        ],
    )


def data_downgrade():
    op.execute("delete from task")
    op.execute("delete from chore")



from abc import ABC, abstractmethod
from datetime import date
from typing import Callable

from pychores.domain.entity.chore import Chore


class ICreateChore(ABC):
    @abstractmethod
    def save_new_chore(self, username: str, chore: Chore) -> Chore:
        """Save a new Chore for username"""


class InvalidChorePayload(Exception):
    pass


class CreateChore:
    def __init__(self, repo: ICreateChore, date_provider: Callable[[], date]):
        self.repo = repo
        self.date_provider = date_provider

    def execute(self, username: str, payload: dict) -> Chore:
        try:
            chore = Chore(
                name=payload["name"],
                description=payload["description"],
                period=payload["period"],
                current_date=self.date_provider(),
                last_execution=None,
            )
        except ValueError:
            raise InvalidChorePayload()
        created_chore = self.repo.save_new_chore(username=username, chore=chore)
        return created_chore
